// math.cpp

#include "jpnslisp.h"

extern CELL *NILCell,*TCell;

long double CellToNumber(CELL *p)
{
    if (p->attr != 'n') return 0.0;
    return ((AFLOAT *)(p->car))->f;
}

CELL *NumberToCell(long double v)
{
    CELL *p;

    p = getCell();
    p->attr = 'n';
    p->car = (CELL *)getNumber(v);
    return p;
}

long CellToInt(CELL *p)
{
    if (p->attr != 'i') return 0;
    return ((AINT *)(p->car))->f;
}

CELL *IntToCell(long v)
{
    CELL *p;

    p = getCell();
    p->attr = 'i';
    p->car = (CELL *)getInt(v);
    return p;
}

long CellToBunshi(CELL *p)
{
    if (p->attr != 'b') return 0;
    return ((AINT *)(p->car))->f;
}

long CellToBunbo(CELL *p)
{
    if (p->attr != 'b') return 0;
    return ((AINT *)(p->cdr))->f;
}

CELL *BunsuToCell(long i,long o)
{
    CELL *p;

    p = getCell();
    p->attr = 'b';
    p->car = (CELL *)getInt(i);
    p->cdr = (CELL *)getInt(o);
    return p;
}

CELL *jSqrt(CELL *p,SYMVAL *loc,long locn,long lock)
{
    long double v;
    long n;
    long vd,vu;
    long i,k;
    BOOL bunshiInt;
    long bunshi;

    p = eval(p,loc,locn,lock);
    if (p == NULL) return NULL;
    if (p->attr == 'n') {
	v = CellToNumber(p);
	if (v < 0.0) {
	    puts("G[: ͕͂Ƃ܂");
	    return NULL;
	}
	v = sqrt((double)v);
	return NumberToCell(v);
    }
    else if (p->attr == 'i') {
	n = CellToInt(p);
	if (n < 0) {
	    puts("G[: ͕͂Ƃ܂");
	    return NULL;
	}
	v = sqrt((double)n);
	vd = (long)floor((double)v);
	vu = (long)ceil((double)v);
	k = vd * vd;
	if (k == n) return IntToCell(vd);
	k = vu * vu;
	if (k == n) return IntToCell(vu);
	return NumberToCell(v);
    }
    else if (p->attr == 'b') {
	n = CellToBunshi(p);
	if (n < 0) {
	    puts("G[: ͕͂Ƃ܂");
	    return NULL;
	}
	v = (long double)sqrt((double)n);
	vd = (long)floor((double)v);
	vu = (long)ceil((double)v);
	k = vd * vd;
	if (k == n) {
	    bunshi = vd;
	    bunshiInt = TRUE;
	    if (bunshi == 0) return IntToCell(0);
	}
	else {
	    k = vu * vu;
	    if (k == n) {
		bunshi = vu;
		bunshiInt = TRUE;
		if (bunshi == 0) return IntToCell(0);
	    }
	    else {
		bunshiInt = FALSE;
	    }
	}

	if (bunshiInt) {
	    n = CellToBunbo(p);
	    if (n <= 0) {
		puts("G[: ͕͂Ƃ܂");
		return NULL;
	    }
	    v = sqrt((double)n);
	    vd = (long)floor((double)v);
	    vu = (long)ceil((double)v);
	    k = vd * vd;
	    if (k == n) return BunsuToCell(bunshi,vd);
	    k = vu * vu;
	    if (k == n) return BunsuToCell(bunshi,vu);
	    return NumberToCell(bunshi / v);
	}
	else {
	    v = sqrt((long double)(CellToBunshi(p) / (long double)CellToBunbo(p)));
	    return NumberToCell(v);
	}
    }
    else {
	puts("G[: lȊO͂̕Ƃ܂");
	return NULL;
    }
}

CELL *njExpt(long double x,CELL *y)
{
    switch (y->attr) {
	case 'n':
	    return NumberToCell(pow(x,CellToNumber(y)));
	case 'i':
	    return NumberToCell(pow(x,(long double)CellToInt(y)));
	case 'b':
	    return NumberToCell(pow(x,(long double)CellToBunshi(y) / (long double)CellToBunbo(y)));
	default:
	    puts("G[: w͐lłȂ΂Ȃ܂");
	    return NULL;
    }
}

CELL *ijExpt(long x,CELL *y)
{
    long i,n,k;
    long double d;
    long dd,du;
    long limit;

    switch (y->attr) {
	case 'n':
	    return NumberToCell(pow(x,CellToNumber(y)));
	case 'i':
	    k = 1;
	    n = CellToInt(y);
	    if (x != 0L) limit = LONG_MAX / x;
	    else limit = LONG_MAX;
	    for (i = 0;i < n;i++) {
		if (k >= limit) {
		    puts("G[: lI[o[t[܂");
		    return NULL;
		}
		k *= x;
	    }
	    return IntToCell(k);
	case 'b':
	    d = pow(x,1.0 / (long double)CellToBunbo(y));
	    dd = (long)floor(d);
	    du = (long)ceil(d);
	    k = 1L;
	    n = CellToBunbo(y);
	    for (i = 0;i < n;i++) k *= dd;
	    if (x == k) {
		k = 1L;
		n = CellToBunshi(y);
		limit = LONG_MAX / dd;
		for (i = 0;i < n;i++) {
		    if (k >= limit) {
			puts("G[: lI[o[t[܂");
			return NULL;
		    }
		    k *= dd;
		}
		return IntToCell(k);
	    }
	    k = 1L;
	    for (i = 0;i < n;i++) k *= du;
	    if (x == k) {
		k = 1L;
		n = CellToBunshi(y);
		limit = LONG_MAX / du;
		for (i = 0;i < n;i++) {
		    if (k >= limit) {
			puts("G[: lI[o[t[܂");
			return NULL;
		    }
		    k *= du;
		}
		return IntToCell(k);
	    }
	    return NumberToCell(pow(x,(long double)CellToBunshi(y) / (long double)CellToBunbo(y)));
	default:
	    puts("G[: w͐lłȂ΂Ȃ܂");
	    return NULL;
    }
}

CELL *bjExpt(long a,long b,CELL *y)
{
    long shi,bo,i,n;
    long double d;
    long dd,du;
    long k,k2;
    long limit;

    switch (y->attr) {
	case 'i':
	    n = CellToInt(y);
	    shi = 1L;
	    if (a != 0L) limit = LONG_MAX / a;
	    else limit = LONG_MAX;
	    for (i = 0L;i < n;i++) {
		if (shi >= limit) { puts("G[: qI[o[t[܂"); return NULL; }
		shi *= a;
	    }
	    bo = 1L;
	    if (b != 0L) limit = LONG_MAX / b;
	    else limit = LONG_MAX;
	    for (i = 0L;i < n;i++) {
		if (bo >= limit) { puts("G[: ꂪI[o[t[܂"); return NULL; }
		bo *= b;
	    }
	    if (bo == 0L) { puts("G[: ꂪOł"); return NULL; }
	    return BunsuToCell(shi,bo);
	case 'n':
	    return NumberToCell(pow((long double)a / (long double)b,CellToNumber(y)));
	case 'b':
	    d = pow((long double)a,1.0 / CellToBunbo(y));
	    dd = (long)floor(d);
	    du = (long)ceil(d);
	    n = CellToBunbo(y);
	    k = 1L;
	    for (i = 0;i < n;i++) k *= dd;
	    k2 = 1L;
	    for (i = 0;i < n;i++) k2 *= du;
	    if (k == a || k2 == a) {
		if (k == a) shi = dd; else shi = du;

		d = pow((long double)b,1.0 / CellToBunbo(y));
		dd = (long)floor(d);
		du = (long)ceil(d);
		n = CellToBunbo(y);
		k = 1L;
		for (i = 0;i < n;i++) k *= dd;
		k2 = 1L;
		for (i = 0;i < n;i++) k2 *= du;
		if (k == b || k2 == b) {
		    if (k == b) bo = dd; else bo = du;
		    k = 1L;
		    n = CellToBunshi(y);
		    for (i = 0;i < n;i++) k *= shi;
		    k2 = 1L;
		    for (i = 0;i < n;i++) k2 *= bo;
		    return BunsuToCell(k,k2);
		}
		else {
		    return NumberToCell(pow((long double)a / (long double)b,
			    (long double)CellToBunshi(y) / (long double)CellToBunbo(y) ) );
	        }
	    }
	    else {
		return NumberToCell(pow((long double)a / (long double)b,
			(long double)CellToBunshi(y) / (long double)CellToBunbo(y) ) );
	    }
	default:
	    puts("G[: w͐lłȂ΂Ȃ܂");
	    return NULL;
    }
}

BOOL isPlus(CELL *p)
{
    if (p->attr == 'n') {
	if (CellToNumber(p) >= 0.0) return TRUE;
	else return FALSE;
    }
    else if (p->attr == 'i') {
	if (CellToInt(p) >= 0) return TRUE;
	else return FALSE;
    }
    else if (p->attr == 'b') {
	if (CellToBunshi(p) >= 0) return TRUE;
	else return FALSE;
    }
    else return FALSE;
}

BOOL isPlusExceptZero(CELL *p)
{
    if (p->attr == 'n') {
	if (CellToNumber(p) > 0.0) return TRUE;
	else return FALSE;
    }
    else if (p->attr == 'i') {
	if (CellToInt(p) > 0) return TRUE;
	else return FALSE;
    }
    else if (p->attr == 'b') {
	if (CellToBunshi(p) > 0) return TRUE;
	else return FALSE;
    }
    else return FALSE;
}

CELL *jAbs(CELL *p)
{
    switch (p->attr) {
	case 'n':
	    return NumberToCell(-1.0 * CellToNumber(p));
	case 'i':
	    return IntToCell(-1L * CellToInt(p));
	case 'b':
	    return BunsuToCell(-1L * CellToBunshi(p),CellToBunbo(p));
	default:
	    puts("G[: w͐lłȂ΂Ȃ܂");
	    return NULL;
    }
}

CELL *jExpt(CELL *x,CELL *y,SYMVAL *loc,long locn,long lock)
{

    IncCellStack(NULL);
    IncCellStack(y);

    x = eval(x,loc,locn,lock);

    PopCellStack();

    if (x == NULL) return NULL;

    IncCellStack(NULL);
    IncCellStack(x);

    y = eval(y,loc,locn,lock);

    PopCellStack();

    if (y == NULL) return NULL;

    if (!isPlus(x)) { puts("G[: Kvł"); return NULL; }

    if (x->attr == 'i' && isPlus(y)) {
	return ijExpt(CellToInt(x),y);
    }
    else if (x->attr == 'n') {
	return njExpt(CellToNumber(x),y);
    }
    else if (x->attr == 'b' && isPlus(y)) {
	return bjExpt(CellToBunshi(x),CellToBunbo(x),y);
    }
    else if (x->attr == 'i') {
	y = jAbs(y);
	if (y == NULL) return NULL;
	return bjExpt(1L,CellToInt(x),y);
    }
    else if (x->attr == 'b' && CellToBunshi(x) == 1L) {
	y = jAbs(y);
	if (y == NULL) return NULL;
	return ijExpt(CellToBunbo(x),y);
    }
    else if (x->attr == 'b') {
	y = jAbs(y);
	if (y == NULL) return NULL;
	return bjExpt(CellToBunbo(x),CellToBunshi(x),y);
    }
    else {
	puts("G[: lׂ̂ł͂܂");
	return NULL;
    }
}

CELL *jExp(CELL *p,SYMVAL *loc,long locn,long lock)
{
    long v;

    p = eval(p,loc,locn,lock);
    if (p == NULL) return NULL;
    if (p->attr == 'i') {
	v = CellToInt(p);
	if (v == 0L) return IntToCell(1L);
	else {
	    return NumberToCell(exp((double)v));
	}
    }
    else if (p->attr == 'n') {
	return NumberToCell(exp(CellToNumber(p)));
    }
    else if (p->attr == 'b') {
	return NumberToCell(exp((long double)CellToBunshi(p) / (long double)CellToBunbo(p)));
    }
    else {
	puts("G[: EXP l̈Kvł");
	return NULL;
    }
}

CELL *jLog(CELL *p,SYMVAL *loc,long locn,long lock)
{
    long v;

    p = eval(p,loc,locn,lock);
    if (p == NULL) return NULL;
    if (!isPlusExceptZero(p)) { puts("G[: Kvł"); return NULL; }
    if (p->attr == 'i') {
	v = CellToInt(p);
	if (v == 1L) return IntToCell(0L);
	else {
	    return NumberToCell(log((double)v));
	}
    }
    else if (p->attr == 'n') {
	return NumberToCell(log(CellToNumber(p)));
    }
    else if (p->attr == 'b') {
	return NumberToCell(log((long double)CellToBunshi(p) / (long double)CellToBunbo(p)));
    }
    else {
	puts("G[: LOG l̈Kvł");
	return NULL;
    }
}

CELL *jSin(CELL *p,SYMVAL *loc,long locn,long lock)
{
    long v;

    p = eval(p,loc,locn,lock);
    if (p == NULL) return NULL;
    if (p->attr == 'i') {
	v = CellToInt(p);
	if (v == 0L) return IntToCell(0L);
	else {
	    return NumberToCell(sin((double)v));
	}
    }
    else if (p->attr == 'n') {
	return NumberToCell(sin(CellToNumber(p)));
    }
    else if (p->attr == 'b') {
	return NumberToCell(sin((long double)CellToBunshi(p) / (long double)CellToBunbo(p)));
    }
    else {
	puts("G[: SIN l̈Kvł");
	return NULL;
    }
}

CELL *jCos(CELL *p,SYMVAL *loc,long locn,long lock)
{
    long v;

    p = eval(p,loc,locn,lock);
    if (p == NULL) return NULL;
    if (p->attr == 'i') {
	v = CellToInt(p);
	if (v == 0L) return IntToCell(1L);
	else {
	    return NumberToCell(cos((double)v));
	}
    }
    else if (p->attr == 'n') {
	return NumberToCell(cos(CellToNumber(p)));
    }
    else if (p->attr == 'b') {
	return NumberToCell(cos((long double)CellToBunshi(p) / (long double)CellToBunbo(p)));
    }
    else {
	puts("G[: COS l̈Kvł");
	return NULL;
    }
}

CELL *jTan(CELL *p,SYMVAL *loc,long locn,long lock)
{
    long v;

    p = eval(p,loc,locn,lock);
    if (p == NULL) return NULL;
    if (p->attr == 'i') {
	v = CellToInt(p);
	if (v == 0L) return IntToCell(0L);
	else {
	    return NumberToCell(tan((double)v));
	}
    }
    else if (p->attr == 'n') {
	return NumberToCell(tan(CellToNumber(p)));
    }
    else if (p->attr == 'b') {
	return NumberToCell(tan((long double)CellToBunshi(p) / (long double)CellToBunbo(p)));
    }
    else {
	puts("G[: TAN l̈Kvł");
	return NULL;
    }
}

BOOL isInRange(CELL *p)
{
    long i;
    long double v;

    switch (p->attr) {
	case 'i':
	    i = CellToInt(p);
	    if (-1L <= i && i <= 1L) return TRUE;
	    else return FALSE;
	case 'n':
	    v = CellToNumber(p);
	    if (-1.0 <= v && v <= 1.0) return TRUE;
	    else return FALSE;
	case 'b':
	    v = (long double)CellToBunshi(p) / (long double)CellToBunbo(p);
	    if (-1.0 <= v && v <= 1.0) return TRUE;
	    else return FALSE;
	default:
	    return FALSE;
    }
}

CELL *jAsin(CELL *p,SYMVAL *loc,long locn,long lock)
{
    long i;
    long double v;

    p = eval(p,loc,locn,lock);
    if (p == NULL) return NULL;
    if (!isInRange(p)) { puts("G[: ͈͊Oł"); return NULL; }
    switch (p->attr) {
	case 'i':
	    i = CellToInt(p);
	    if (i == 0L) return IntToCell(0L);
	    return NumberToCell(asin((long double)i));
	case 'n':
	    return NumberToCell(asin(CellToNumber(p)));
	case 'b':
	    v = (long double)CellToBunshi(p) / (long double)CellToBunbo(p);
	    return NumberToCell(asin(v));
	default:
	    puts("G[: lł͂܂");
	    return NULL;
    }
}

CELL *jAcos(CELL *p,SYMVAL *loc,long locn,long lock)
{
    long i;
    long double v;

    p = eval(p,loc,locn,lock);
    if (p == NULL) return NULL;
    if (!isInRange(p)) { puts("G[: ͈͊Oł"); return NULL; }
    switch (p->attr) {
	case 'i':
	    i = CellToInt(p);
	    if (i == 1L) return IntToCell(0L);
	    return NumberToCell(acos((long double)i));
	case 'n':
	    return NumberToCell(acos(CellToNumber(p)));
	case 'b':
	    v = (long double)CellToBunshi(p) / (long double)CellToBunbo(p);
	    return NumberToCell(acos(v));
	default:
	    puts("G[: lł͂܂");
	    return NULL;
    }
}

CELL *jAtan(CELL *p,SYMVAL *loc,long locn,long lock)
{
    long i;
    long double v;

    p = eval(p,loc,locn,lock);
    if (p == NULL) return NULL;
    switch (p->attr) {
	case 'i':
	    i = CellToInt(p);
	    if (i == 0L) return IntToCell(0L);
	    return NumberToCell(atan((long double)i));
	case 'n':
	    return NumberToCell(atan(CellToNumber(p)));
	case 'b':
	    v = (long double)CellToBunshi(p) / (long double)CellToBunbo(p);
	    return NumberToCell(atan(v));
	default:
	    puts("G[: lł͂܂");
	    return NULL;
    }
}

CELL *jEvenp(CELL *p,SYMVAL *loc,long locn,long lock)
{
    long v,w,x;

    p = eval(p,loc,locn,lock);
    if (p == NULL) return NULL;
    if (p->attr != 'i') { puts("G[: ł͂܂"); return NULL; }
    v = CellToInt(p);
    w = v / 2L;
    x = v - w * 2L;
    if (x == 1L || x == -1L) return NILCell;
    return TCell;
}

CELL *tfloor(CELL *p,SYMVAL *loc,long locn,long lock)
{
	long double f;
	AINT *k;
	long k1,k2;
	CELL *q,*a,*r,*r2,*q2;
	AFLOAT *n;

	p = eval(p,loc,locn,lock);
	if (p == NULL) return NULL;
	if (p->attr == 'i') {
	    q = getCell();
	    *q = *p;
	    k = getInt(0L);
	    a = getCell();
	    a->attr = 'i';
	    a->car = (CELL *)k;
	    r = getCell();
	    r2 = getCell();
	    r->attr = 'l';
	    r->car = q;
	    r->cdr = r2;
	    r2->attr = 'l';
	    r2->car = a;
	    r2->cdr = NILCell;
	    q->mtop = r;
	    return q;
	}
	else if (p->attr == 'n') {
	    q = getCell();
	    k1 = (long)floor((double)(((AFLOAT *)(p->car))->f));
	    k = getInt(k1);
	    q->attr = 'i';
	    q->car = (CELL *)k;
	    f = ((AFLOAT *)(p->car))->f - k1;
	    n = getNumber(f);
	    a = getCell();
	    a->attr = 'n';
	    a->car = (CELL *)n;
	    r = getCell();
	    r2 = getCell();
	    r->attr = 'l';
	    r->car = q;
	    r->cdr = r2;
	    r2->attr = 'l';
	    r2->car = a;
	    r2->cdr = NILCell;
	    q->mtop = r;
	    return q;
	}
	else if (p->attr == 'b') {
	    q = getCell();
	    k1 = CellToBunshi(p);
	    k2 = CellToBunbo(p);
	    f = (long double)k1 / (long double)k2;
	    k1 = (long)floor((double)f);
	    k = getInt(k1);
	    q->attr = 'i';
	    q->car = (CELL *)k;
	    q2 = getCell();
	    *q2 = *q;
	    a = Sa(p,q2);
	    r = getCell();
	    r2 = getCell();
	    r->attr = 'l';
	    r->car = q;
	    r->cdr = r2;
	    r2->attr = 'l';
	    r2->car = a;
	    r2->cdr = NILCell;
	    q->mtop = r;
	    return q;
	}
	else {
	    puts("G[: lł͂܂");
	    return NULL;
	}
}

CELL *now(void)
{
    time_t t;
    CELL *w,*ca,*cd;
    AINT *ia,*id;

    t = time(NULL);
    w = getCell();
    ca = getCell();
    cd = getCell();
    ia = getInt((long)(t / 3600L));
    id = getInt((long)(t % 3600L));
    w->attr = 'l';
    ca->attr = 'i';
    cd->attr = 'i';
    ca->car = (CELL *)ia;
    cd->car = (CELL *)id;
    w->car = ca;
    w->cdr = cd;
    return w;
}
